from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade import UpgradeStep
from ftw.upgrade.executioner import Executioner
from ftw.upgrade.indexing import HAS_INDEXING
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName
from unittest import skipIf
from zope.component import queryAdapter
from zope.interface.verify import verifyClass
from Products.CMFPlone.utils import getFSVersionTuple

import transaction


class TestExecutioner(UpgradeTestCase):

    def test_component_is_registered(self):
        setup_tool = getToolByName(self.layer['portal'], 'portal_setup')
        executioner = queryAdapter(setup_tool, IExecutioner)
        self.assertNotEqual(None, executioner)

    def test_implements_interface(self):
        verifyClass(IExecutioner, Executioner)

    def test_installs_upgrades(self):
        def upgrade(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            portal.upgrade_step_executed = True

        self.package.with_profile(Builder('genericsetup profile')
                                   .with_upgrade(Builder('plone upgrade step')
                                                 .upgrading('1000', to='1002')
                                                 .calling(upgrade)))

        self.portal.upgrade_step_executed = False
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.assertFalse(self.portal.upgrade_step_executed)
            self.install_profile_upgrades('the.package:default')
            self.assertTrue(self.portal.upgrade_step_executed)

    def test_install_upgrades_by_api_ids(self):
        def upgrade(setup_context):
            portal = setup_context.portal_url.getPortalObject()
            portal.upgrade_step_executed = True

        self.package.with_profile(Builder('genericsetup profile')
                                   .with_upgrade(Builder('plone upgrade step')
                                                 .upgrading('1000', to='1002')
                                                 .calling(upgrade)))

        self.portal.upgrade_step_executed = False
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.assertFalse(self.portal.upgrade_step_executed)
            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids('1002@the.package:default')
            self.assertTrue(self.portal.upgrade_step_executed)

    def test_transaction_note(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled(u'Register "foo" utility'))
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1001', to='1002')
                          .titled(u'Update email address'))
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1002', to='1003')
                          .titled(u'Update email from name')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.install_profile_upgrades('the.package:default')
            self.assertMultiLineEqual(
                u'the.package:default -> 1001 (Register "foo" utility)\n'
                u'the.package:default -> 1002 (Update email address)\n'
                u'the.package:default -> 1003 (Update email from name)',
                transaction.get().description)

    def test_after_commit_hook(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled(u'Register "foo" utility')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.install_profile_upgrades('the.package:default')

            hooks = list(transaction.get().getAfterCommitHooks())
            hook_funcs = [h[0] for h in hooks]

            self.assertIn(
                'notification_hook',
                [f.__name__ for f in hook_funcs],
                'Our notification_hook should be registered')

            transaction.commit()
            self.assertEqual(
                [],
                list(transaction.get().getAfterCommitHooks()),
                'Hook registrations should not persist across transactions')

    def test_resources_are_recooked_after_installing_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1000', to='1001')))
        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            with self.assert_resources_recooked():
                self.install_profile_upgrades('the.package:default')

    @skipIf(getFSVersionTuple() > (5, 1), 'QuickInstaller has been deprecated in Plone 5.1')
    def test_updates_quickinstaller_version(self):
        quickinstaller = getToolByName(self.portal, 'portal_quickinstaller')

        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001')))
        self.package.with_version('1.1')

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            quickinstaller.get('the.package').installedversion = '1.0'
            self.assertEqual('1.0', quickinstaller.get('the.package').getInstalledVersion())
            self.install_profile_upgrades('the.package:default')
            self.assertEqual('1.1', quickinstaller.get('the.package').getInstalledVersion())

    def test_install_profiles_by_profile_ids(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001')))

        self.setup_logging()
        profile_id = 'the.package:default'
        with self.package_created():
            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_profiles_by_profile_ids(profile_id)
            self.assertEqual(
                ['Installing profile the.package:default.',
                 'Done installing profile the.package:default.'],
                self.get_log())

            self.purge_log()
            executioner.install_profiles_by_profile_ids(profile_id)
            self.assertEqual(
                ['Ignoring already installed profile the.package:default.'],
                self.get_log())

            self.purge_log()
            executioner.install_profiles_by_profile_ids(profile_id,
                                                        force_reinstall=True)
            self.assertEqual(
                ['Installing profile the.package:default.',
                 'Done installing profile the.package:default.'],
                self.get_log())

    def test_do_not_decrease_version_when_only_installing_orphan_steps(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 11, 11, 11, 11)))
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2012, 12, 12, 12, 12))))

        with self.package_created():
            self.install_profile('the.package:default', '20121212121200')
            self.clear_recorded_upgrades('the.package:default')
            self.record_installed_upgrades('the.package:default', '20121212121200')
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['20121212121200'],
                    'proposed': ['20111111111100'],
                    'orphan': ['20111111111100']}})
            self.assertEqual(
                (u'20121212121200',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))

            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids(
                '20111111111100@the.package:default')
            self.assertEqual(
                (u'20121212121200',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['20111111111100', '20121212121200'],
                    'proposed': [],
                    'orphan': []}})

    @skipIf(not HAS_INDEXING,
            'Tests must only run when indexing is available')
    def test_logs_indexing_progress_of_final_reindex(self):
        self.grant('Manager')
        create(Builder('folder'))
        create(Builder('folder'))

        class TriggerReindex(UpgradeStep):
            def __call__(self):
                catalog = self.getToolByName("portal_catalog")
                for brain in catalog(portal_type="Folder"):
                    brain.getObject().reindexObject()

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2013, 1, 1))
                          .calling(TriggerReindex)))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            self.setup_logging()

            self.install_profile_upgrades('the.package:default')
            self.assertEqual(
                '1 of 2 (50%): Processing indexing queue',
                self.get_log()[-1])

    def test_regression_switching_versioning_system(self):
        # test_do_not_decrease_version_when_only_installing_orphan_steps
        # caused an issue when switching to upgrade step directories,
        # since the versioning system changed and '4' > '2016' but 4 < 2016.

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('4000', to='4002'))
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 11, 11, 11, 11))))

        with self.package_created():
            self.install_profile('the.package:default', '4002')
            self.clear_recorded_upgrades('the.package:default')
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['4002'],
                    'proposed': ['20111111111100'],
                    'orphan': []}})
            self.assertEqual(
                (u'4002',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))

            executioner = queryAdapter(self.portal_setup, IExecutioner)
            executioner.install_upgrades_by_api_ids(
                '20111111111100@the.package:default')
            self.assertEqual(
                (u'20111111111100',),
                self.portal_setup.getLastVersionForProfile('the.package:default'))
            self.assert_gathered_upgrades({
                'the.package:default': {
                    'done': ['4002', '20111111111100'],
                    'proposed': [],
                    'orphan': []}})



from datetime import datetime
from DateTime import DateTime
from ftw.builder import Builder
from ftw.builder import create
from ftw.builder.content import dx_content_builders_registered
from ftw.testing import freeze
from ftw.upgrade.migration import BACKUP_AND_IGNORE_UNMAPPED_FIELDS
from ftw.upgrade.migration import DISABLE_FIELD_AUTOMAPPING
from ftw.upgrade.migration import FieldsNotMappedError
from ftw.upgrade.migration import IBaseObject
from ftw.upgrade.migration import IGNORE_DEFAULT_IGNORE_FIELDS
from ftw.upgrade.migration import IGNORE_STANDARD_FIELD_MAPPING
from ftw.upgrade.migration import IGNORE_UNMAPPED_FIELDS
from ftw.upgrade.migration import InplaceMigrator
from ftw.upgrade.migration import UNMAPPED_FIELDS_BACKUP_ANN_KEY
from ftw.upgrade.tests.base import UpgradeTestCase
from operator import attrgetter
from plone.app.relationfield.behavior import IRelatedItems
from plone.app.testing import login
from plone.app.testing import SITE_OWNER_NAME
from plone.app.textfield import RichTextValue
from plone.dexterity.interfaces import IDexterityContent
from plone.portlets.interfaces import IPortletAssignmentMapping
from plone.portlets.interfaces import IPortletManager
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces.constrains import ENABLED
from Products.CMFPlone.interfaces.constrains import IConstrainTypes
from Products.CMFPlone.interfaces.constrains import ISelectableConstrainTypes
from Products.CMFPlone.utils import getFSVersionTuple
from six.moves import map
from unittest import skipIf
from zope.annotation import IAnnotations
from zope.component import getMultiAdapter
from zope.component import getUtility


def todt(date):
    if isinstance(date, DateTime):
        return todt(date.asdatetime())
    return date.replace(tzinfo=None)


@skipIf(getFSVersionTuple() > (5, ),
        'The inplace migrator migrates from Archetypes.'
        ' Plone 5 has no Archetypes objects.')
class TestInplaceMigrator(UpgradeTestCase):

    def setUp(self):
        super(TestInplaceMigrator, self).setUp()
        self.wftool = getToolByName(self.portal, 'portal_workflow')
        self.wftool.setChainForPortalTypes(
            ['Document', 'Folder'],
            'simple_publication_workflow')

    def test_migrate_archetypes_folder_to_dexterity(self):
        self.grant('Manager')

        creation_date = datetime(2015, 11, 29, 10, 45)
        modification_date = datetime(2016, 1, 2, 9, 30)
        effective_date = datetime(2016, 2, 2, 2, 30)
        expires_date = datetime(2016, 3, 3, 3, 30)

        with freeze(creation_date):
            folder = create(Builder('folder')
                            .titled(u'The Folder')
                            .having(description='The Description',
                                    excludeFromNav=True,
                                    subject='One\nTwo',
                                    effectiveDate=DateTime(effective_date),
                                    expirationDate=DateTime(expires_date))
                            .in_state('published'))

        setattr(folder, '__ac_local_roles_block__', True)
        create(Builder('user').named('Hugo', 'Boss')
               .with_roles('Reader', 'Editor', on=self.portal))  # local roles!
        create(Builder('user').named('John', 'Doe')
               .with_roles('Reader', 'Editor', on=folder))

        folder._setProperty('search_label', 'HB', 'string')

        with freeze(modification_date):
            folder.reindexObject()  # update modification date

        self.assertTrue(IBaseObject.providedBy(folder))
        self.assertFalse(IDexterityContent.providedBy(folder))
        self.assertEqual(('', 'plone', 'the-folder'), folder.getPhysicalPath())
        self.assertEqual('The Folder', folder.Title())
        self.assertEqual('The Description', folder.Description())
        self.assertEqual(True, folder.exclude_from_nav())
        self.assertEqual(('One', 'Two'), folder.Subject())
        self.assertEqual(todt(creation_date), todt(folder.created()))
        self.assertEqual(todt(modification_date), todt(folder.modified()))
        self.assertEqual(todt(effective_date), todt(folder.effective()))
        self.assertEqual(todt(expires_date), todt(folder.expires()))
        self.assertEqual('HB', folder.getProperty('search_label', None))
        self.assertEqual('published', self.review_state(folder))
        self.assertEqual((('john.doe', ('Reader', 'Editor')),
                          ('test_user_1_', ('Owner',))),
                         folder.get_local_roles())

        old_catalog_indexdata = self.get_catalog_indexdata_for(folder)

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('Folder').migrate_object(folder)

        folder = self.portal.get('the-folder')
        self.assertFalse(IBaseObject.providedBy(folder))
        self.assertTrue(IDexterityContent.providedBy(folder))
        self.assertEqual(('', 'plone', 'the-folder'), folder.getPhysicalPath())
        self.assertEqual('The Folder', folder.Title())
        self.assertEqual('The Description', folder.Description())
        self.assertEqual(True, folder.exclude_from_nav)
        self.assertEqual(('One', 'Two'), folder.Subject())
        self.assertEqual(todt(creation_date), todt(folder.created()))
        self.assertEqual(todt(modification_date), todt(folder.modified()))
        self.assertEqual(todt(effective_date), todt(folder.effective()))
        self.assertEqual(todt(expires_date), todt(folder.expires()))
        self.assertEqual('HB', folder.getProperty('search_label', None))
        self.assertEqual('published', self.review_state(folder))
        self.assertEqual((('john.doe', ('Reader', 'Editor')),
                          ('test_user_1_', ('Owner',))),
                         folder.get_local_roles())

        self.maxDiff = None

        new_catalog_indexdata = self.get_catalog_indexdata_for(folder)
        # In Plone 4.3.7, the SearchableText changes here.
        del old_catalog_indexdata['SearchableText']
        del new_catalog_indexdata['SearchableText']

        self.assertDictEqual(old_catalog_indexdata, new_catalog_indexdata)

    def test_migrate_dexterity_folder_to_dexterity(self):
        self.grant('Manager')
        self.install_profile('plone.app.contenttypes:default')

        creation_date = datetime(2015, 11, 29, 10, 45)
        modification_date = datetime(2016, 1, 2, 9, 30)
        effective_date = datetime(2016, 2, 2, 2, 30)
        expires_date = datetime(2016, 3, 3, 3, 30)

        with dx_content_builders_registered():
            with freeze(creation_date):
                folder = create(Builder('folder')
                                .titled(u'The Folder')
                                .having(description=u'The Description',
                                        exclude_from_nav=True,
                                        subjects=(u'One', u'Two'),
                                        effective=effective_date,
                                        expires=expires_date)
                                .in_state('pending'))

        with freeze(modification_date):
            folder.reindexObject()  # update modification date

        self.assertFalse(IBaseObject.providedBy(folder))
        self.assertTrue(IDexterityContent.providedBy(folder))
        self.assertEqual(('', 'plone', 'the-folder'), folder.getPhysicalPath())
        self.assertEqual('The Folder', folder.Title())
        self.assertEqual('The Description', folder.Description())
        self.assertEqual(True, folder.exclude_from_nav)
        self.assertEqual(('One', 'Two'), folder.Subject())
        self.assertEqual(todt(creation_date), todt(folder.created()))
        self.assertEqual(todt(modification_date), todt(folder.modified()))
        self.assertEqual(todt(effective_date), todt(folder.effective()))
        self.assertEqual(todt(expires_date), todt(folder.expires()))
        self.assertEqual('pending', self.review_state(folder))

        old_catalog_indexdata = self.get_catalog_indexdata_for(folder)

        InplaceMigrator('Folder').migrate_object(folder)

        folder = self.portal.get('the-folder')
        self.assertFalse(IBaseObject.providedBy(folder))
        self.assertTrue(IDexterityContent.providedBy(folder))
        self.assertEqual(('', 'plone', 'the-folder'), folder.getPhysicalPath())
        self.assertEqual('The Folder', folder.Title())
        self.assertEqual('The Description', folder.Description())
        self.assertEqual(True, folder.exclude_from_nav)
        self.assertEqual(('One', 'Two'), folder.Subject())
        self.assertEqual(todt(creation_date), todt(folder.created()))
        self.assertEqual(todt(modification_date), todt(folder.modified()))
        self.assertEqual(todt(effective_date), todt(folder.effective()))
        self.assertEqual(todt(expires_date), todt(folder.expires()))
        self.assertEqual('pending', self.review_state(folder))

        self.maxDiff = None
        self.assertDictEqual(old_catalog_indexdata,
                             self.get_catalog_indexdata_for(folder))

    def test_migrate_archetypes_page_to_dexterity(self):
        self.grant('Manager')

        page = create(Builder('page')
                      .titled(u'The Page')
                      .having(text='<p>Some Text</p>')
                      .in_state('published'))

        self.assertTrue(IBaseObject.providedBy(page))
        self.assertFalse(IDexterityContent.providedBy(page))
        self.assertEqual('The Page', page.Title())
        self.assertEqual('<p>Some Text</p>', page.getText())
        self.assertEqual('published', self.review_state(page))

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('Document').migrate_object(page)

        page = self.portal.get('the-page')
        self.assertFalse(IBaseObject.providedBy(page))
        self.assertTrue(IDexterityContent.providedBy(page))
        self.assertEqual(('', 'plone', 'the-page'), page.getPhysicalPath())
        self.assertEqual('The Page', page.Title())
        self.assertIsInstance(page.text, RichTextValue)
        self.assertEqual('<p>Some Text</p>', page.text.output)
        self.assertEqual('published', self.review_state(page))

    def test_migrate_empty_richtextfield_to_dexterity(self):
        self.grant('Manager')

        no_text_page = create(Builder('page')
                              .titled(u'No Text Page')
                              .having(text=u'')
                              .in_state('published'))

        self.assertFalse(IDexterityContent.providedBy(no_text_page))
        self.assertTrue(IBaseObject.providedBy(no_text_page))
        self.assertEqual(u'', no_text_page.getText())

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('Document').migrate_object(no_text_page)

        no_text_page = self.portal.get('no-text-page')
        self.assertFalse(IBaseObject.providedBy(no_text_page))
        self.assertTrue(IDexterityContent.providedBy(no_text_page))
        self.assertNotIsInstance(no_text_page.text, RichTextValue)

    def test_migrate_archetypes_file_to_dexterity(self):
        self.grant('Manager')

        thefile = create(
            Builder('file')
            .titled(u'The File')
            .attach_file_containing('<doc>Content</doc>', name='data.xml'))

        self.assertTrue(IBaseObject.providedBy(thefile))
        self.assertFalse(IDexterityContent.providedBy(thefile))
        self.assertEqual('The File', thefile.Title())

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('File').migrate_object(thefile)

        thefile = self.portal.get('the-file')
        self.assertFalse(IBaseObject.providedBy(thefile))
        self.assertTrue(IDexterityContent.providedBy(thefile))
        self.assertEqual(('', 'plone', 'the-file'), thefile.getPhysicalPath())
        self.assertEqual('The File', thefile.Title())
        self.assertEqual('data.xml', thefile.file.filename)
        self.assertEqual('<doc>Content</doc>', thefile.file.data)

    def test_field_mapping_overrides_auto_mapping(self):
        self.grant('Manager')
        folder = create(Builder('folder').titled(u'The Title'))
        self.install_profile('plone.app.contenttypes:default')

        InplaceMigrator('Folder', {'title': 'description'},
                        ignore_fields=['description']).migrate_object(folder)
        folder = self.portal.get(folder.getId())
        self.assertEqual(u'The Title', folder.Description())

    def test_DISABLE_FIELD_AUTOMAPPING_flag(self):
        """When disabling field automapping we expect unmaped fields.
        """

        self.grant('Manager')
        folder = create(Builder('folder'))
        self.install_profile('plone.app.contenttypes:default')

        with self.assertRaises(FieldsNotMappedError) as cm:
            (InplaceMigrator('Folder', options=DISABLE_FIELD_AUTOMAPPING)
             .migrate_object(folder))

        self.assertIn('title', cm.exception.not_mapped_fields)

    def test_IGNORE_UNMAPPED_FIELDS_flag(self):
        self.grant('Manager')
        folder = create(Builder('folder'))
        self.install_profile('plone.app.contenttypes:default')

        (InplaceMigrator(
            'Folder',
            options=DISABLE_FIELD_AUTOMAPPING | IGNORE_UNMAPPED_FIELDS)
         .migrate_object(folder))

    def test_BACKUP_AND_IGNORE_UNMAPPED_FIELDS_flag(self):
        self.grant('Manager')
        folder = create(Builder('folder')
                        .titled(u'The Folder')
                        .having(description='A very fancy folder.'))
        self.install_profile('plone.app.contenttypes:default')

        new_folder = (
            InplaceMigrator(
                'Folder',
                options=DISABLE_FIELD_AUTOMAPPING | BACKUP_AND_IGNORE_UNMAPPED_FIELDS)
            .migrate_object(folder))

        self.assertEqual(
            {'nextPreviousEnabled': False,
             'description': u'A very fancy folder.',
             'contributors': (),
             'title': u'The Folder',
             'rights': u'',
             'language': u'en',
             'relatedItems': [],
             'creators': (u'test_user_1_',)},
            IAnnotations(new_folder).get(
                UNMAPPED_FIELDS_BACKUP_ANN_KEY, None))

    def test_IGNORE_STANDARD_FIELD_MAPPING_flag(self):
        self.grant('Manager')
        folder = create(Builder('folder'))
        self.install_profile('plone.app.contenttypes:default')

        with self.assertRaises(FieldsNotMappedError) as cm:
            (InplaceMigrator('Folder',
                             options=IGNORE_STANDARD_FIELD_MAPPING)
             .migrate_object(folder))

        self.assertIn('expirationDate', cm.exception.not_mapped_fields)

    def test_IGNORE_DEFAULT_IGNORE_FIELDS_flag(self):
        self.grant('Manager')
        folder = create(Builder('folder'))
        self.install_profile('plone.app.contenttypes:default')

        with self.assertRaises(FieldsNotMappedError) as cm:
            (InplaceMigrator('Folder',
                             options=IGNORE_DEFAULT_IGNORE_FIELDS)
             .migrate_object(folder))

        self.assertIn('location', cm.exception.not_mapped_fields)

    def test_migrate_constrain_types(self):
        self.grant('Manager')
        self.maxDiff = None

        folder = create(Builder('folder').titled(u'The Folder'))
        self.set_constraintypes_config(
            folder,
            {'mode': ENABLED,
             'locally allowed': ['Folder', 'Document', 'File'],
             'immediately addable': ['Folder', 'Document']})

        self.assertTrue(IBaseObject.providedBy(folder))
        self.assertFalse(IDexterityContent.providedBy(folder))
        self.assertDictEqual(
            {'mode': ENABLED,
             'locally allowed': {'Folder', 'Document', 'File'},
             'immediately addable': {'Folder', 'Document'}},
            self.get_constraintypes_config(folder))

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('Folder').migrate_object(folder)

        folder = self.portal.get('the-folder')
        self.assertFalse(IBaseObject.providedBy(folder))
        self.assertTrue(IDexterityContent.providedBy(folder))
        self.assertDictEqual(
            {'mode': ENABLED,
             'locally allowed': {'Folder', 'Document', 'File'},
             'immediately addable': {'Folder', 'Document'}},
            self.get_constraintypes_config(folder))

    def test_migrate_ownership(self):
        john = create(Builder('user').named('John', 'Doe').with_roles('Manager'))
        peter = create(Builder('user').named('Peter', 'Pan').with_roles('Manager'))

        login(self.portal, john.getId())
        folder = create(Builder('folder').titled(u'The Folder'))
        folder.changeOwnership(peter.getUser())

        self.assertTrue(IBaseObject.providedBy(folder))
        self.assertEqual('john.doe', folder.Creator())
        self.assertEqual('peter.pan', folder.getOwner().getId())

        self.grant('Manager')
        self.install_profile('plone.app.contenttypes:default')
        with self.login(SITE_OWNER_NAME):
            InplaceMigrator('Folder').migrate_object(folder)

        folder = self.portal.get('the-folder')
        self.assertTrue(IDexterityContent.providedBy(folder))
        self.assertEqual('john.doe', folder.Creator())
        self.assertEqual('peter.pan', folder.getOwner().getId())

    def test_migrate_ownership_no_IOwner(self):
        john = create(Builder('user').named('John', 'Doe').with_roles('Manager'))
        peter = create(Builder('user').named('Peter', 'Pan').with_roles('Manager'))

        login(self.portal, john.getId())
        folder = create(Builder('folder').titled(u'The Folder'))
        folder.changeOwnership(peter.getUser())

        self.assertTrue(IBaseObject.providedBy(folder))
        self.assertEqual('john.doe', folder.Creator())
        self.assertEqual('peter.pan', folder.getOwner().getId())

        self.grant('Manager')
        self.install_profile('plone.app.contenttypes:default')

        # The creators list behaves differently when the dublin core
        # behavior is used.
        self.portal.portal_types['Folder'].behaviors = tuple(
            name for name in self.portal.portal_types['Folder'].behaviors
            if not (name == 'plone.dublincore' or 'IDublinCore' in name
                    or name == 'plone.ownership'or 'IOwnership' in name))
        with self.login(SITE_OWNER_NAME):
            InplaceMigrator('Folder', options=IGNORE_UNMAPPED_FIELDS).migrate_object(folder)

        folder = self.portal.get('the-folder')
        self.assertTrue(IDexterityContent.providedBy(folder))
        self.assertEqual('john.doe', folder.Creator())
        self.assertEqual('peter.pan', folder.getOwner().getId())

    def test_migrate_object_position(self):
        self.grant('Manager')
        container = create(Builder('folder').titled(u'Container'))
        one = create(Builder('folder').titled(u'One').within(container))
        two = create(Builder('folder').titled(u'Two').within(container))
        three = create(Builder('folder').titled(u'Three').within(container))

        self.assertEqual(
            [0, 1, 2],
            list(map(container.getObjectPosition, ('one', 'two', 'three'))))
        container.moveObjectsByDelta(['three'], -1)
        self.assertEqual(
            [0, 2, 1],
            list(map(container.getObjectPosition, ('one', 'two', 'three'))))

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('Folder').migrate_object(container)
        InplaceMigrator('Folder').migrate_object(three)
        InplaceMigrator('Folder').migrate_object(two)
        InplaceMigrator('Folder').migrate_object(one)
        container = self.portal.get('container')

        self.assertEqual(
            [0, 2, 1],
            list(map(container.getObjectPosition, ('one', 'two', 'three'))))

    def test_migrate_portlets(self):
        self.grant('Manager')

        folder = create(Builder('folder').titled(u'The Folder'))
        portlet = create(Builder('static portlet')
                         .within(folder)
                         .in_manager('plone.rightcolumn'))

        self.assertEqual({'plone.leftcolumn': [],
                          'plone.rightcolumn': [portlet]},
                         self.get_portlets_for(folder))

        self.install_profile('plone.app.contenttypes:default')
        InplaceMigrator('Folder').migrate_object(folder)

        folder = self.portal.get('the-folder')
        self.assertEqual({'plone.leftcolumn': [],
                          'plone.rightcolumn': [portlet]},
                         self.get_portlets_for(folder))

    def test_migrate_relations(self):
        self.grant('Manager')

        foo = create(Builder('folder').titled(u'Foo'))
        bar = create(Builder('folder').titled(u'Bar')
                     .having(relatedItems=[foo]))

        self.assertEqual([foo], bar.getRelatedItems())
        self.assertEqual([bar], foo.getBackReferences())

        self.install_profile('plone.app.contenttypes:default')
        list(map(InplaceMigrator('Folder').migrate_object, (foo, bar)))

        foo = self.portal.get('foo')
        bar = self.portal.get('bar')

        self.assertEqual(
            [foo],
            list(map(attrgetter('to_object'), IRelatedItems(bar).relatedItems)))

    def get_catalog_indexdata_for(self, obj):
        catalog = getToolByName(obj, 'portal_catalog')
        uid = '/'.join(obj.getPhysicalPath())
        data = catalog.getIndexDataForUID(uid)

        # Remove some indexes in order to not assert them
        # since they must change from AT to DX in
        del data['meta_type']
        del data['object_provides']
        del data['created']
        del data['modified']
        data.pop('sync_uid', None)
        return data

    def review_state(self, obj):
        return self.wftool.getInfoFor(obj, 'review_state', None)

    def get_constraintypes_config(self, obj):
        ctypes = IConstrainTypes(obj)
        return {'mode': ctypes.getConstrainTypesMode(),
                'locally allowed': set(ctypes.getLocallyAllowedTypes()),
                'immediately addable': set(
                    ctypes.getImmediatelyAddableTypes())}

    def set_constraintypes_config(self, obj, config):
        self.assertEqual(
            {'mode', 'locally allowed', 'immediately addable'},
            set(config))

        ctypes = ISelectableConstrainTypes(obj)
        ctypes.setConstrainTypesMode(config['mode'])
        ctypes.setLocallyAllowedTypes(config['locally allowed'])
        ctypes.setImmediatelyAddableTypes(config['immediately addable'])

    def get_portlets_for(self, container):
        portlets = {}

        for manager_name in ('plone.leftcolumn', 'plone.rightcolumn'):
            manager = getUtility(
                IPortletManager,
                name=manager_name,
                context=self.portal
            )
            assignments = getMultiAdapter(
                (container, manager),
                IPortletAssignmentMapping,
                context=self.portal
            )
            portlets[manager_name] = list(assignments.values())

        return portlets



from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Combine JS/CSS bundles together (Plone 5).
[quote]
    $ ./bin/upgrade combine_bundles --site Plone
[/quote]
""".format(
    t=TERMINAL
).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'combine_bundles', help='Combine JS/CSS bundles together (Plone 5).', description=DOCS
    )
    command.set_defaults(func=combine_bundles)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)


@with_api_requestor
@error_handling
def combine_bundles(args, requestor):
    print((requestor.POST('combine_bundles').json()))



from contextlib import contextmanager
from datetime import datetime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.directory.scanner import Scanner
from ftw.upgrade.exceptions import UpgradeStepDefinitionError
from ftw.upgrade.tests.base import UpgradeTestCase
from six.moves import map


class TestDirectoryScanner(UpgradeTestCase):

    def setUp(self):
        super(TestDirectoryScanner, self).setUp()
        self.profile = Builder('genericsetup profile')
        self.package.with_profile(self.profile)

    def test_returns_chained_upgrade_infos(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add an action'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 2, 2, 8))
                                  .named('update the action'))
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 3, 3, 8))
                                  .named('remove the action'))

        with self.scanned() as upgrade_infos:
            list(map(lambda info: (info.__delitem__('path'),
                                   info.__delitem__('callable')),
                     upgrade_infos))

            self.maxDiff = None
            self.assertEqual(
                [{'source-version': None,
                  'target-version': '20110101080000',
                  'title': 'Add an action.'},

                 {'source-version': '20110101080000',
                  'target-version': '20110202080000',
                  'title': 'Update the action.'},

                 {'source-version': '20110202080000',
                  'target-version': '20110303080000',
                  'title': 'Remove the action.'}],

                upgrade_infos)

    def test_exception_raised_when_upgrade_has_no_code(self):
        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add action')
                                  .with_code(''))

        with create(self.package) as package:
            with self.assertRaises(UpgradeStepDefinitionError) as cm:
                self.scan(package)

        self.assertEqual(
            'The upgrade step 20110101080000_add_action has no upgrade class'
            ' in the upgrade.py module.',
            str(cm.exception))

    def test_exception_raised_when_multiple_upgrade_steps_detected(self):
        code = '\n'.join((
                'from ftw.upgrade import UpgradeStep',
                'class Foo(UpgradeStep): pass',
                'class Bar(UpgradeStep): pass'))

        self.profile.with_upgrade(Builder('ftw upgrade step')
                                  .to(datetime(2011, 1, 1, 8))
                                  .named('add action')
                                  .with_code(code))

        with create(self.package) as package:
            with self.assertRaises(UpgradeStepDefinitionError) as cm:
                self.scan(package)

        self.assertEqual(
            'The upgrade step 20110101080000_add_action has more than one upgrade'
            ' class in the upgrade.py module.',
            str(cm.exception))

    def test_does_not_fail_when_no_upgrades_present(self):
        self.package.with_zcml_include('ftw.upgrade', file='meta.zcml')
        self.package.with_zcml_node('upgrade-step:directory',
                                    profile='the.package:default',
                                    directory='.')

        with self.scanned() as upgrade_infos:
            self.assertEqual( [], upgrade_infos)

    @contextmanager
    def scanned(self):
        with create(self.package) as package:
            yield self.scan(package)

    def scan(self, package):
        upgrades = package.package_path.joinpath('upgrades')
        return Scanner('the.package.upgrades', upgrades).scan()



from ftw.builder import Builder
from ftw.builder import create
from ftw.builder.testing import BUILDER_LAYER
from ftw.testing.layer import TempDirectoryLayer
from ftw.upgrade.command.utils import find_egginfo
from ftw.upgrade.command.utils import find_package_namespace_path
from ftw.upgrade.tests.helpers import capture_streams
from ftw.upgrade.tests.helpers import chdir
from path import Path
from six import StringIO
from unittest import TestCase


LAYER = TempDirectoryLayer(bases=(BUILDER_LAYER, ),
                           name='ftw.upgrade.test_command_utils')


class TestFindEgginfo(TestCase):
    layer = LAYER

    def setUp(self):
        self.path = Path(self.layer['temp_directory']).realpath()
        self.package_builder = (Builder('python package')
                                .named('the.package')
                                .at_path(self.path))

    def test_finds_egginfo_in_current_directory(self):
        create(self.package_builder)
        self.assertEqual(self.path.joinpath('the.package.egg-info'),
                         find_egginfo(self.path))

    def test_working_directory_used_when_no_path_passed_to_in(self):
        create(self.package_builder)
        with chdir(self.path):
            self.assertEqual(self.path.joinpath('the.package.egg-info'),
                             find_egginfo())

    def test_prints_warning_when_no_egginfo_found(self):
        stderr = StringIO()
        with capture_streams(stderr=stderr):
            self.assertEqual(None, find_egginfo(self.path))

        self.assertEqual('WARNING: no *.egg-info directory could be found.\n',
                         stderr.getvalue())

    def test_prints_warning_when_multiple_egginfos_found(self):
        create(self.package_builder)
        create(Builder('python package').named('another.package').at_path(self.path))

        stderr = StringIO()
        with capture_streams(stderr=stderr):
            self.assertEqual(None, find_egginfo(self.path))

        self.assertEqual('WARNING: more than one *.egg-info directory found.\n',
                         stderr.getvalue())


class TestFindPackageNamespacePath(TestCase):
    layer = LAYER

    def setUp(self):
        self.path = Path(self.layer['temp_directory']).realpath()

    def test_returns_absolute_path_to_toplevel_namespace_directory(self):
        create(Builder('python package').named('the.package').at_path(self.path))
        egginfo = find_egginfo(self.path)
        self.assertEqual(self.path.joinpath('the'),
                         find_package_namespace_path(egginfo))



from AccessControl.SecurityInfo import ClassSecurityInformation
from Acquisition import aq_base
from Acquisition import aq_parent
from ftw.upgrade.events import ClassMigratedEvent
from ftw.upgrade.exceptions import NoAssociatedProfileError
from ftw.upgrade.helpers import update_security_for
from ftw.upgrade.interfaces import IUpgradeStep
from ftw.upgrade.progresslogger import ProgressLogger
from ftw.upgrade.utils import log_silencer
from ftw.upgrade.utils import SavepointIterator
from ftw.upgrade.utils import SizedGenerator
from plone.browserlayer.interfaces import ILocalBrowserLayerType
from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletManagerRenderer
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2Base
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.utils import getToolByName
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zExceptions import NotFound
from zope.browser.interfaces import IBrowserView
from zope.event import notify
from zope.interface import directlyProvidedBy
from zope.interface import directlyProvides
from zope.interface import implementer
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest

import logging
import re
import six


try:
    from Products.GenericSetup.tool import DEPENDENCY_STRATEGY_NEW
except ImportError:
    DEPENDENCY_STRATEGY_NEW = None

try:
    from Products.CMFPlone.utils import get_installer
except ImportError:
    get_installer = None


LOG = logging.getLogger('ftw.upgrade')


@implementer(IUpgradeStep)
class UpgradeStep(object):
    security = ClassSecurityInformation()

    deferrable = False

    def __new__(cls, *args, **kwargs):
        """Let the class act as function since we cannot registry a
        classmethod directly.
        This does call the object immediately after instantiating.
        """
        obj = object.__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj()

    def __init__(self, portal_setup,
                 associated_profile=None,
                 base_profile=None,
                 target_version=None):
        self.portal_setup = portal_setup
        self.portal = self.getToolByName('portal_url').getPortalObject()
        self.associated_profile = associated_profile
        self.base_profile = base_profile
        self.target_version = target_version

    security.declarePrivate('__call__')
    def __call__(self):
        """This method is implemented in each upgrade step with the
        tasks the upgrade should perform.
        """
        raise NotImplementedError()

    security.declarePrivate('getToolByName')
    def getToolByName(self, tool_name):
        """Returns the tool with the name ``tool_name`` of the upgraded
        site.
        """
        return getToolByName(self.portal_setup, tool_name)

    security.declarePrivate('objects')
    def objects(self, catalog_query, message, logger=None,
                savepoints=None):
        """Queries the catalog (unrestricted) and an iterator with full
        objects.
        The iterator configures and calls a ``ProgressLogger`` with the
        passed ``message``.
        """

        objects = self.catalog_unrestricted_search(
            catalog_query, full_objects=True)

        objects = SavepointIterator.build(objects, savepoints, logger)
        return ProgressLogger(message, objects, logger=logger)

    security.declarePrivate('catalog_rebuild_index')
    def catalog_rebuild_index(self, name):
        """Reindex the ``portal_catalog`` index identified by ``name``.
        """
        catalog = self.getToolByName('portal_catalog')
        LOG.info("Reindexing index %s" % name)

        # pylint: disable=W0212
        pgthreshold = catalog._getProgressThreshold() or 100
        # pylint: enable=W0212
        pghandler = ZLogHandler(pgthreshold)
        catalog.reindexIndex(name, None, pghandler=pghandler)

        LOG.info("Reindexing index %s DONE" % name)

    security.declarePrivate('catalog_reindex_objects')
    def catalog_reindex_objects(self, query, idxs=None, savepoints=None):
        """Reindex all objects found in the catalog with `query`.
        A list of indexes can be passed as `idxs` for limiting the
        indexed indexes.
        """

        title = '.'.join((self.__module__, self.__class__.__name__))

        for obj in self.objects(query, title, savepoints=savepoints):
            if idxs is None:
                # Store modification date
                modification_date = obj.modified()
                obj.reindexObject()

                # Restore modification date
                obj.setModificationDate(modification_date)
                obj.reindexObject(idxs=['modified'])

            else:
                obj.reindexObject(idxs=idxs)

    security.declarePrivate('catalog_has_index')
    def catalog_has_index(self, name):
        """Returns whether there is a catalog index ``name``.
        """
        catalog = self.getToolByName('portal_catalog')
        index_names = catalog.indexes()
        return name in index_names

    security.declarePrivate('catalog_add_index')
    def catalog_add_index(self, name, type_, extra=None):
        """Adds a new index to the ``portal_catalog`` tool.
        """
        catalog = self.getToolByName('portal_catalog')
        return catalog.addIndex(name, type_, extra=extra)

    security.declarePrivate('catalog_remove_index')
    def catalog_remove_index(self, name):
        """Removes an index to from ``portal_catalog`` tool.
        """
        catalog = self.getToolByName('portal_catalog')
        return catalog.delIndex(name)

    security.declarePrivate('catalog_unrestricted_get_object')
    def catalog_unrestricted_get_object(self, brain):
        """Returns the unrestricted object of a brain.
        """
        try:
            return self.portal.unrestrictedTraverse(brain.getPath())
        except (AttributeError, KeyError, NotFound):
            LOG.warning('The object of the brain with rid {!r} no longer'
                        ' exists at the path {!r}; removing the brain.'.format(
                            brain.getRID(), brain.getPath()))
            catalog = self.getToolByName('portal_catalog')
            catalog.uncatalog_object(brain.getPath())
            return None

    security.declarePrivate('catalog_unrestricted_search')
    def catalog_unrestricted_search(self, query, full_objects=False):
        """Search catalog without security checks.
        If `full_objects` is `True`, objects instead of brains
        are returned.
        """
        catalog = self.getToolByName('portal_catalog')
        brains = tuple(catalog.unrestrictedSearchResults(query))

        if full_objects:
            generator = (self.catalog_unrestricted_get_object(brain)
                         for brain in brains)
            generator = (obj for obj in generator if obj is not None)
            return SizedGenerator(generator, len(brains))

        else:
            return brains

    security.declarePrivate('actions_remove_action')
    def actions_remove_action(self, category, action_id):
        """Removes an action identified by ``action_id`` from
        the ``portal_actions`` tool from a particulary ``category``.
        """

        actions_tool = self.getToolByName('portal_actions')
        cat = actions_tool.get(category)

        if cat and action_id in cat:
            del cat[action_id]
            return True

        else:
            return False

    security.declarePrivate('actions_remove_type_action')
    def actions_remove_type_action(self, portal_type, action_id):
        """Removes a ``portal_types`` action from the type identified
        by ``portal_type`` with the action id ``action_id``.
        """
        ttool = self.getToolByName('portal_types')
        fti = ttool.get(portal_type)

        actions = []
        found = False

        for action in fti._actions:  # pylint: disable=W0212
            if action.id != action_id:
                actions.append(action)
            else:
                found = True

        fti._actions = tuple(actions)  # pylint: disable=W0212
        return found

    security.declarePrivate('actions_add_type_action')
    def actions_add_type_action(self, portal_type, after, action_id, **kwargs):
        """Add a ``portal_types`` action from the type identified
        by ``portal_type``, the position can be defined by the
        ``after`` attribute. If the after action can not be found,
        the action will be inserted at the end of the list."""

        actions = []
        found = False

        ttool = self.getToolByName('portal_types')
        fti = ttool.get(portal_type)

        new_action = ActionInformation(id=action_id, **kwargs)

        for action in fti._actions:  # pylint: disable=W0212
            actions.append(action)
            if action.id == after:
                actions.append(new_action)
                found = True

        if not found:
            actions.append(new_action)

        fti._actions = tuple(actions)  # pylint: disable=W0212

    security.declarePrivate('set_property')
    def set_property(self, context, key, value, data_type='string'):
        """Set a property with the key ``value`` and the value ``value``
        on the ``context`` safely. The property is created with the
        type ``data_type`` if it does not exist.
        """

        # pylint: disable=W0212
        if context.hasProperty(key):
            context._updateProperty(key, value)
        else:
            context._setProperty(key, value, data_type)
        # pylint: enable=W0212

    security.declarePrivate('add_lines_to_property')
    def add_lines_to_property(self, context, key, lines):
        """Updates a property with key ``key`` on the object ``context``
        adding ``lines``. The property is expected to be of type "lines".
        If the property does not exist it is created.
        """

        if context.hasProperty(key):
            data = list(getattr(context, key))

            if isinstance(lines, (list, tuple)):
                data.extend(lines)
            else:
                data.append(lines)

        elif isinstance(lines, (list, tuple)):
            data = lines

        else:
            data = [lines]

        self.set_property(context, key, data, data_type='lines')

    security.declarePrivate('setup_install_profile')
    def setup_install_profile(self, profileid, steps=None,
                              dependency_strategy=None):
        """Installs the generic setup profile identified by ``profileid``.
        If a list step names is passed with ``steps`` (e.g. ['actions']),
        only those steps are installed. All steps are installed by default.
        """
        setup = self.getToolByName('portal_setup')
        if steps is None:
            runargs = {'purge_old': False}
            if DEPENDENCY_STRATEGY_NEW is not None:
                runargs['dependency_strategy'] = (
                    dependency_strategy or DEPENDENCY_STRATEGY_NEW)

            setup.runAllImportStepsFromProfile(profileid, **runargs)
        else:
            for step in steps:
                setup.runImportStepFromProfile(profileid,
                                               step,
                                               run_dependencies=False,
                                               purge_old=False)

    security.declarePrivate('ensure_profile_installed')
    def ensure_profile_installed(self, profileid):
        """Install a generic setup profile only when it is not yet installed.
        """
        if not self.is_profile_installed(profileid):
            self.setup_install_profile(profileid)

    security.declarePrivate('install_upgrade_profile')
    def install_upgrade_profile(self, steps=None):
        """Installs the generic setup profile for this upgrade step.
        """
        if self.associated_profile is None:
            raise NoAssociatedProfileError()

        self.setup_install_profile(self.associated_profile, steps=steps)

    security.declarePrivate('is_product_installed')
    def is_profile_installed(self, profileid):
        """Checks whether a generic setup profile is installed.
        Respects product uninstallation via quickinstaller.
        """
        profileid = re.sub(r'^profile-', '', profileid)

        try:
            profileinfo = self.portal_setup.getProfileInfo(profileid)
        except KeyError:
            return False

        if not self.is_product_installed(profileinfo['product']):
            return False

        version = self.portal_setup.getLastVersionForProfile(profileid)
        return version != 'unknown'

    security.declarePrivate('is_product_installed')
    def is_product_installed(self, product_name):
        """Check whether a product is installed.
        """
        if get_installer is not None:
            quickinstaller = get_installer(self.portal, self.portal.REQUEST)
            return (quickinstaller.is_product_installable(product_name)
                    and quickinstaller.is_product_installed(product_name))
        else:
            quickinstaller = self.getToolByName('portal_quickinstaller')
            return (quickinstaller.isProductInstallable(product_name)
                    and quickinstaller.isProductInstalled(product_name))


    security.declarePrivate('uninstall_product')
    def uninstall_product(self, product_name):
        """Uninstalls a product using the quick installer.
        """
        if get_installer is not None:
            quickinstaller = get_installer(self.portal, self.portal.REQUEST)
            quickinstaller.uninstall_product(product_name)
        else:
            quickinstaller = self.getToolByName('portal_quickinstaller')
            quickinstaller.uninstallProducts([product_name])

    security.declarePrivate('migrate_class')
    def migrate_class(self, obj, new_class):
        """Changes the class of a object and notifies the container so that
        the change is persistent.
        It has a special handling for BTreeFolder2Base based containers.

        Fires an event after class migration so that custom reference cleanup
        can be performed.
        """
        obj.__class__ = new_class

        base = aq_base(obj)
        base._ofs_migrated = True
        base._p_changed = True

        parent = aq_base(aq_parent(obj))
        id_ = base.getId()

        if isinstance(parent, BTreeFolder2Base):
            del parent._tree[id_]
            parent._tree[id_] = base

        else:
            parent._p_changed = True

        # Refresh provided interfaces cache
        directlyProvides(base, directlyProvidedBy(base))

        notify(ClassMigratedEvent(obj))

    security.declarePrivate('remove_broken_browserlayer')
    def remove_broken_browserlayer(self, name, dottedname):
        """Removes a browser layer registration, whose interface can't be
        imported any more, from the persistent registry.
        """
        iface_name = dottedname.split('.')[-1]
        sm = self.portal.getSiteManager()
        adapters = sm.utilities._adapters
        subscribers = sm.utilities._subscribers

        # Remove adapters ...
        if name in adapters[0][ILocalBrowserLayerType]:
            del adapters[0][ILocalBrowserLayerType][name]

        # ... as well as subscribers
        layer_subscribers = subscribers[0][ILocalBrowserLayerType]
        remaining_layers = tuple([layer for layer in layer_subscribers['']
                                if not layer.__name__ == iface_name])
        layer_subscribers[''] = remaining_layers

        sm._utility_registrations.pop((ILocalBrowserLayerType, name), None)

        sm.utilities._p_changed = True

    security.declarePrivate('remove_broken_portlet_manager')
    def remove_broken_portlet_manager(self, name):
        """Removes a portlet manager, that cannot be imported any more, from
        the persistent registry.

        This method also filters some pretty ugly error messages from the
        zodb connection log. The errors would appear when zodb attempts to load
        the objects for classes that have been uninstalled previously.
        """
        sm = self.portal.getSiteManager()

        manager_renderer = sm.adapters.lookup(
            [Interface, IBrowserRequest, IBrowserView],
            IPortletManagerRenderer,
            name)
        if manager_renderer is not None:
            with log_silencer("ZODB.Connection", "Couldn't load state for"):
                sm.unregisterAdapter(
                    manager_renderer,
                    [Interface, IBrowserRequest, IBrowserView],
                    IPortletManagerRenderer,
                    name)
            LOG.info("Removed portlet manager renderer {0}".format(name))

        with log_silencer("ZODB.Connection", "Couldn't load state for"):
            manager = sm.queryUtility(IPortletManager, name=name)
            if manager is not None:
                sm.unregisterUtility(component=manager,
                                     name=name,
                                     provided=IPortletManager)
                LOG.info("Removed portlet manager {0}".format(name))

    security.declarePrivate('update_security')
    def update_security(self, obj, reindex_security=True):
        """Update the object security and reindex the security indexes in
        the catalog.
        """
        return update_security_for(obj, reindex_security=reindex_security)

    security.declarePrivate('update_workflow_security')
    def update_workflow_security(self, workflow_names, reindex_security=True,
                                 savepoints=1000):
        """Updates the object security of all objects with one of the
        passed workflows.
        `workflows` is expected to be a list of workflow names.
        If `savepoints` is None, no savepoints will be created.
        """

        if getattr(workflow_names, '__iter__', None) is None or \
                isinstance(workflow_names, (str, six.text_type)):
            raise ValueError(
                '"workflows" must be a list of workflow names.')

        from ftw.upgrade.workflow import WorkflowSecurityUpdater
        updater = WorkflowSecurityUpdater()
        updater.update(workflow_names, reindex_security=reindex_security,
                       savepoints=savepoints)



from ftw.upgrade.resource_registries import recook_resources
from ftw.upgrade.tests.base import UpgradeTestCase


class TestResourceRegistries(UpgradeTestCase):

    def test_recooking_resources(self):
        with self.assert_resources_recooked():
            recook_resources()



from BTrees.OOBTree import OOBTree
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.annotation import IAnnotations
from zope.component import adapts
from zope.interface import implementer
from zope.interface import Interface

import six


ANNOTATION_KEY = 'ftw.upgrade:recorder'


@implementer(IUpgradeStepRecorder)
class UpgradeStepRecorder(object):
    adapts(IPloneSiteRoot, Interface)

    def __init__(self, portal, profilename):
        self.portal = portal
        self.profile = self._normalize_profilename(profilename)

    def is_installed(self, target_version):
        storage = self._get_profile_storage()
        return storage and bool(storage.get(target_version, False))

    def mark_as_installed(self, target_version):
        storage = self._get_profile_storage(create=True)
        storage[target_version] = True

    def clear(self):
        self._get_profile_storage(create=True).clear()

    def _get_profile_storage(self, create=False):
        annotations = IAnnotations(self.portal)
        if ANNOTATION_KEY not in annotations and not create:
            return None

        if ANNOTATION_KEY not in annotations:
            annotations[ANNOTATION_KEY] = OOBTree()

        if self.profile not in annotations[ANNOTATION_KEY]:
            annotations[ANNOTATION_KEY][self.profile] = OOBTree()

        return annotations[ANNOTATION_KEY][self.profile]

    def _normalize_profilename(self, profilename):
        if profilename.startswith('profile-'):
            profilename = profilename[len('profile-'):]

        if not isinstance(profilename, six.text_type):
            profilename = profilename.decode('utf-8')

        return profilename



from ftw.upgrade.interfaces import IRecordableHandler
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementedBy


def wrap_upgrade_step(handler, upgrade_profile, base_profile, target_version):
    def upgrade_step_wrapper(portal_setup):
        result = handler(portal_setup, upgrade_profile,
                         base_profile=u'profile-' + base_profile,
                         target_version=target_version)

        portal = getToolByName(portal_setup, 'portal_url').getPortalObject()
        recorder = getMultiAdapter((portal, base_profile),
                                   IUpgradeStepRecorder)
        recorder.mark_as_installed(target_version)
        return result
    alsoProvides(upgrade_step_wrapper, IRecordableHandler)
    alsoProvides(upgrade_step_wrapper, implementedBy(handler))
    upgrade_step_wrapper.handler = handler
    return upgrade_step_wrapper



from contextlib import contextmanager

import logging
import os
import sys


@contextmanager
def capture_streams(stdout=None, stderr=None):
    ori_stdout = sys.stdout
    ori_stderr = sys.stderr

    if stdout is not None:
        sys.stdout = stdout
    if stderr is not None:
        sys.stderr = stderr

    try:
        yield
    finally:
        if stdout is not None:
            sys.stdout = ori_stdout
        if stderr is not None:
            sys.stderr = ori_stderr


@contextmanager
def chdir(path):
    before = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(before)


@contextmanager
def verbose_logging():
    original_level = logging.root.getEffectiveLevel()
    logging.root.setLevel(logging.INFO)
    try:
        yield

    finally:
        logging.root.setLevel(original_level)


@contextmanager
def no_logging_threads():
    """In testing, when a request is executed (e.g. from requests) to the ZSERVER
    layer and errors are logged, the logging module might write to the logging
    handler in a thread after the request is finished.
    Since the testrunner tracks left over threads the logging thread is detected
    as leftover thread. Since this thread is a dummy thread it cannot be joined,
    which results in a error in the threading module (python 2.7 bug).
    The result is that from this test on the test log is spammed with threading
    errors for each following test.
    In order to mitigate this issue this context manager temporarily disables
    threading of logging in general so that logs are processed synchronously and
    no left over threads are created.
    """
    original_log_threads = logging.logThreads
    logging.logThreads = 0
    try:
        yield
    finally:
        logging.logThreads = original_log_threads



from ftw.testing import MockTestCase
from ftw.testing.layer import TEMP_DIRECTORY
from ftw.upgrade.exceptions import CyclicDependencies
from ftw.upgrade.utils import find_cyclic_dependencies
from ftw.upgrade.utils import format_duration
from ftw.upgrade.utils import get_sorted_profile_ids
from ftw.upgrade.utils import get_tempfile_authentication_directory
from ftw.upgrade.utils import SizedGenerator
from ftw.upgrade.utils import subject_from_docstring
from ftw.upgrade.utils import topological_sort
from six.moves import map
from six.moves import range
from unittest import TestCase

import six
import stat


class TestTopologicalSort(TestCase):

    def test_simple(self):
        items = ['b', 'a', 'c']
        dependencies = (
            ('a', 'b'),
            ('b', 'c'))

        self.assertEqual(['a', 'b', 'c'],
                         topological_sort(items, dependencies))

    def test_advanced(self):
        items = ['a', 'c', 'b', 'd']
        dependencies = (
            ('a', 'b'),
            ('a', 'c'),
            ('b', 'd'),
            ('b', 'c'))

        self.assertEqual(['a', 'b', 'c', 'd'],
                         topological_sort(items, dependencies))

    def test_duplicated(self):
        items = ['a', 'b', 'a']
        dependencies = (
            ('b', 'a'),
            )

        self.assertEqual(['b', 'a'],
                         topological_sort(items, dependencies))

    def test_cyclic(self):
        items = ['a', 'b']
        dependencies = (
            ('a', 'b'),
            ('b', 'a'))

        self.assertEqual(None,
                         topological_sort(items, dependencies))

    def test_root_nodes_are_reversed_ordered(self):
        items = ['policy', 'xy', 'foo']

        self.assertEqual(
            ['xy', 'policy', 'foo'],
            topological_sort(items, (('policy', 'foo'),)))

        self.assertEqual(
            ['xy', 'policy', 'foo'],
            topological_sort(reversed(items), (('policy', 'foo'),)),
            'items input order should not change result order')


class TestFindCyclicDependencies(TestCase):

    def test_direct_cyclic_dependencies(self):
        dependencies = (
            ('bar', 'foo'),
            ('baz', 'foo'),
            ('baz', 'bar'),
            ('foo', 'bar'),
            ('default', 'file-replacement'),
            ('default', 'image-replacement'),
            )

        self.assertEqual(
            [set(('foo', 'bar'))],
            list(map(set, find_cyclic_dependencies(dependencies))))

    def test_indirect_cyclic_dependencies(self):
        dependencies = (
            ('foo', 'bar'),
            ('bar', 'baz'),
            ('baz', 'foo'),
            )

        self.assertEqual(
            [set(('foo', 'bar', 'baz'))],
            list(map(set, find_cyclic_dependencies(dependencies))))


class TestSizedGenerator(TestCase):

    def test_length(self):
        generator = SizedGenerator((i for i in range(3)), 3)
        self.assertEqual(3, len(generator))

    def test_iterating(self):
        generator = SizedGenerator((i for i in range(3)), 3)
        self.assertEqual([0, 1, 2], list(generator))


class TestSortedProfileIds(MockTestCase):

    def test_dependencies_resolved(self):
        listProfileInfo = self.mock()
        listProfileInfo.return_value = [
            {'id': 'baz',

             'dependencies': [
                 'profile-foo',
                 'profile-bar']},
            {'id': 'foo'},
            {'id': 'bar',
             'dependencies': ['profile-foo']}
        ]
        portal_setup = self.create_dummy()
        portal_setup.listProfileInfo = listProfileInfo

        self.assertEqual(
            ['foo', 'bar', 'baz'],
            get_sorted_profile_ids(portal_setup)
        )

    def test_root_profiles_are_ordered_by_profile_name(self):
        """In the this example the profiles "baz" and "xy"
        have no dependencies to another and thus might be
        ordered in any order from the graph point of view.
        However, we want a cosistent ordern and therefore
        order those root nodes by name.
        """
        listProfileInfo = self.mock()
        listProfileInfo.return_value = [
            {'id': 'baz',
             'dependencies': ['profile-foo']},
            {'id': 'foo'},
            {'id': 'xy'}
        ]
        portal_setup = self.create_dummy()
        portal_setup.listProfileInfo = listProfileInfo

        self.assertEqual(
            ['foo', 'baz', 'xy'],
            get_sorted_profile_ids(portal_setup))

    def test_cyclic_dependencies(self):
        listProfileInfo = self.mock()
        listProfileInfo.return_value = [
            {'id': 'foo',
             'dependencies': ['profile-bar']},
            {'id': 'bar',
             'dependencies': ['profile-foo']},
            {'id': 'baz',
             'dependencies': []}
        ]
        portal_setup = self.create_dummy()
        portal_setup.listProfileInfo = listProfileInfo

        with self.assertRaises(CyclicDependencies) as cm:
            get_sorted_profile_ids(portal_setup)

        six.assertCountEqual(
            self, ('foo', 'bar'), cm.exception.cyclic_dependencies[0])

        self.assertEqual([('foo', 'bar'), ('bar', 'foo')],
                         cm.exception.dependencies)


class TestFormatDuration(TestCase):

    def test_zero_seconds_is_supported(self):
        self.assertEqual('0 seconds', format_duration(0))

    def test_single_second_is_singular(self):
        self.assertEqual('1 second', format_duration(1))

    def test_multiple_seconds_is_plural(self):
        self.assertEqual('2 seconds', format_duration(2))

    def test_single_minute_is_singular(self):
        self.assertEqual(['1 minute',
                          '1 minute, 1 second',
                          '1 minute, 2 seconds'],

                         [format_duration(60),
                          format_duration(60 + 1),
                          format_duration(60 + 2)])

    def test_multiple_minutes_is_plural(self):
        self.assertEqual(['2 minutes, 1 second',
                          '2 minutes, 2 seconds'],

                         [format_duration((2 * 60) + 1),
                          format_duration((2 * 60) + 2)])

    def test_single_hour_is_singular(self):
        self.assertEqual(['1 hour',
                          '1 hour, 1 minute',
                          '1 hour, 2 minutes, 1 second',
                          '1 hour, 2 minutes, 2 seconds'],

                         [format_duration((60 * 60)),
                          format_duration((60 * 60) + 60),
                          format_duration((60 * 60) + (2 * 60) + 1),
                          format_duration((60 * 60) + (2 * 60) + 2)])

    def test_multiple_hours_is_plural(self):
        self.assertEqual(['2 hours',
                          '2 hours, 1 minute',
                          '2 hours, 2 minutes, 1 second',
                          '2 hours, 2 minutes, 2 seconds'],

                         [format_duration((2 * 60 * 60)),
                          format_duration((2 * 60 * 60) + 60),
                          format_duration((2 * 60 * 60) + (2 * 60) + 1),
                          format_duration((2 * 60 * 60) + (2 * 60) + 2)])

    def test_floating_point_seconds_are_ceiled(self):
        self.assertEqual(['1 second',
                          '1 second',
                          '2 seconds',
                          '2 seconds'],

                         [format_duration(0.1),
                          format_duration(0.9),
                          format_duration(1.1),
                          format_duration(1.9)])


class TestSubjectFromDocstring(TestCase):

    def test_one_line_only(self):
        self.assertEqual(
            'This is the subject.',
            subject_from_docstring('This is the subject.'))

    def test_whitespace_is_stripped(self):
        self.assertEqual(
            'This is the subject.',
            subject_from_docstring('\nThis is the subject.\n'))

    def test_only_subject_is_returned(self):
        self.assertEqual(
            'This is the subject.',
            subject_from_docstring('This is the subject.\n\nThis is the body.'))

    def test_multiline_subject_is_joined(self):
        self.assertEqual(
            'This is a subject, with multiple lines.',
            subject_from_docstring('This is a subject,\n'
                                   'with multiple lines.\n'
                                   '\n'
                                   'And the body.'))


class TestGetTempfileAuthenticationDirectory(TestCase):
    layer = TEMP_DIRECTORY

    def setUp(self):
        self.buildoutdir = self.layer['temp_directory']
        self.buildoutdir.joinpath('bin').mkdir()
        self.buildoutdir.joinpath('bin', 'buildout').write_bytes(b'')
        self.buildoutdir.joinpath('var').mkdir()

    def test_directory_is_created(self):
        tmpdir = get_tempfile_authentication_directory(self.buildoutdir)
        self.assertTrue(tmpdir.isdir())

    def test_supports_setguid_flag(self):
        tmpdir = get_tempfile_authentication_directory(self.buildoutdir)
        tmpdir.chmod(tmpdir.stat().st_mode | stat.S_ISGID)
        get_tempfile_authentication_directory(self.buildoutdir)

    def test_detects_unwanted_others_permissions(self):
        tmpdir = get_tempfile_authentication_directory(self.buildoutdir)
        tmpdir.chmod(tmpdir.stat().st_mode | stat.S_IROTH)
        with self.assertRaises(ValueError):
            get_tempfile_authentication_directory(self.buildoutdir)



from __future__ import print_function
from ftw.upgrade.command.jsonapi import add_json_argument
from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL

import six


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Discover Plone sites on this Zope instance and list them.

{t.bold}EXAMPLES:{t.normal}
[quote]
    $ bin/upgrade sites
    $ bin/upgrade sites --auth admin:admin
    $ bin/upgrade sites --auth admin:admin --json
[/quote]
""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'sites',
        help='Discover Plone sites.',
        description=DOCS)
    command.set_defaults(func=sites_command)
    add_requestor_authentication_argument(command)
    add_json_argument(command)


@with_api_requestor
@error_handling
def sites_command(args, requestor):
    response = requestor.GET('list_plone_sites')

    if args.json:
        print(response.text)
    else:
        for site in response.json():
            print(site['path'].ljust(20), six.ensure_str(site['title']))



from datetime import datetime
from ftw.builder import Builder
from ftw.upgrade import UpgradeStep
from ftw.upgrade.directory.subscribers import no_upgrade_step_marking
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter


class TestDirectoryUpgradeSteps(UpgradeTestCase):

    def setUp(self):
        super(TestDirectoryUpgradeSteps, self).setUp()
        self.portal_actions = getToolByName(self.portal, 'portal_actions')

    def test_installing_profile_sets_version_to_latest_upgrade(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2014, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default')
            version = self.portal_setup.getLastVersionForProfile('the.package:default')
            self.assertEqual(('20140101000000',), version)

    def test_installing_profile_marks_all_upgrade_steps_as_installed_too(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2011, 1, 1))))

        with self.package_created():
            recorder = getMultiAdapter((self.portal, 'the.package:default'),
                                       IUpgradeStepRecorder)
            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' before importing profile.')

            self.install_profile('the.package:default')

            self.assertTrue(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to be marked as installed'
                ' after importing profile.')

    def test_installing_profile_does_not_mark_upgrade_step_as_installed_on_partial_import(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2011, 1, 1))))

        with self.package_created():
            recorder = getMultiAdapter((self.portal, 'the.package:default'),
                                       IUpgradeStepRecorder)
            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' before importing profile.')

            self.portal_setup.runImportStepFromProfile('profile-the.package:default', 'typeinfo')

            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' because of a partial import.')

    def test_context_manager_for_disabling_upgrade_step_marking(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2011, 1, 1))))

        with self.package_created():
            recorder = getMultiAdapter((self.portal, 'the.package:default'),
                                       IUpgradeStepRecorder)
            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' before importing profile.')

            with no_upgrade_step_marking():
                self.install_profile('the.package:default')

            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' because marking is prevented for all profiles.')

            with no_upgrade_step_marking('the.package:default'):
                self.install_profile('the.package:default')

            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' because marking is prevented for this profile.')

            with no_upgrade_step_marking('OTHER.package:default'):
                self.install_profile('the.package:default')

            self.assertTrue(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to be marked as installed'
                ' because it is only prevented for other profiles.')

    def test_associated_upgrade_step_profile_is_imported(self):
        class AddTestAction(UpgradeStep):
            def __call__(self):
                self.install_upgrade_profile()

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .with_file('actions.xml', self.asset('test-action.xml'))
                          .calling(AddTestAction)))

        with self.package_created():
            self.install_profile('the.package:default')
            self.assertIsNone(self.get_action())
            self.install_profile_upgrades('the.package:default')
            self.assertEqual('The Test Action', self.get_action().title)

    def test_running_upgrade_steps_marks_them_as_installed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2011, 1, 1))))

        with self.package_created():
            recorder = getMultiAdapter((self.portal, 'the.package:default'),
                                       IUpgradeStepRecorder)

            self.portal_setup.setLastVersionForProfile('the.package:default', u'0')
            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' before importing profile.')

            executioner = IExecutioner(self.portal_setup)
            first_upgrade, = self.portal_setup.listUpgrades('the.package:default')
            executioner.install([('the.package:default', [first_upgrade['id']])])

            self.assertTrue(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to be marked as installed'
                ' after upgrading.')

    def test_base_profile_and_target_version_are_provided(self):
        self.portal.upgrade_infos = {}

        class StoreUpgradeInfos(UpgradeStep):
            def __call__(self):
                self.portal.upgrade_infos.update({
                    'base_profile': self.base_profile,
                    'target_version': self.target_version,
                })

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .calling(StoreUpgradeInfos)))

        with self.package_created():
            self.install_profile('the.package:default')
            self.assertEqual({}, self.portal.upgrade_infos)
            self.install_profile_upgrades('the.package:default')
            self.assertEqual(
                {'base_profile': u'profile-the.package:default',
                 'target_version': '20110101000000'},
                self.portal.upgrade_infos)

    def get_action(self):
        return self.portal_actions.portal_tabs.get('test-action')



from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.helpers import update_security_for
from ftw.upgrade.tests.base import WorkflowTestCase
from Products.CMFPlone.utils import getFSVersionTuple


ALLOWED_ROLES_AND_USERS_PERMISSION = 'View'
if getFSVersionTuple() > (5, 2):
    ALLOWED_ROLES_AND_USERS_PERMISSION = 'Access contents information'


class TestUpdateSecurity(WorkflowTestCase):

    def test_removes_rules_unmanaged_by_workflow(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder'))

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_not_acquired('Modify portal content', folder)
        self.assert_permission_acquired('List folder contents', folder)

        folder.manage_permission('Modify portal content', roles=[],
                                 acquire=True)
        folder.manage_permission('List folder contents', roles=[],
                                 acquire=False)

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_acquired('Modify portal content', folder)
        self.assert_permission_not_acquired('List folder contents', folder)

        update_security_for(folder)

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_not_acquired('Modify portal content', folder)
        self.assert_permission_acquired('List folder contents', folder)

    def test_reindexes_security_indexes(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')
        folder = create(Builder('folder')
                        .in_state('published'))

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users(folder))
        folder.reindexObjectSecurity()

        folder.manage_permission(
            ALLOWED_ROLES_AND_USERS_PERMISSION, roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users(folder))

        update_security_for(folder)

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users(folder))

    def test_without_reindexing_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')
        folder = create(Builder('folder')
                        .in_state('published'))

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users(folder))

        folder.manage_permission(
            ALLOWED_ROLES_AND_USERS_PERMISSION, roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users(folder))

        update_security_for(folder, reindex_security=False)

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users(folder))



from ftw.upgrade.interfaces import IClassMigratedEvent
from zope.component.interfaces import ObjectEvent
from zope.interface import implementer


@implementer(IClassMigratedEvent)
class ClassMigratedEvent(ObjectEvent):
    pass
